package Impedance::H;
#$Id: H.pm,v 1.2 2015/08/13 07:29:14 sofc Exp $
#####################################################################
#
# Package Impedance::H;
#
# This module implements Impedance::Base for diffusion type elements
# (Specifically the Havriliak–Negami element).
#
# To obtain an instance call one of the constructors:
#
# $id = Impedance::H->new($nr);
#
# The module defines no additional member functions.
#
#####################################################################
use strict;
use warnings;
use Impedance::Base;
use Math::Complex;
our @ISA = qw(Impedance::Base);

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    %EXPORT_TAGS = (
        'standard'=>[],
                    );
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
}

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'H';
    $self->{'_tags'} = ['dC','t','a','b'];
    $self->{'_desc'}{'name'} = "Havriliak–Negami";
    $self->{'_desc'}{'dC'} = 'Capacitance diference';
    $self->{'_desc'}{'t'} = 'Relaxation time';
    $self->{'_desc'}{'a'} = 'Assymetry exponent';
    $self->{'_desc'}{'b'} = 'Broadness exponent';
    $self->{'_f_imax'} = 1e-100;
    $self->{'_values'}{'a'} = 0.9;
    $self->{'_values'}{'b'} = 0.9;
    $self->{'_values'}{'dC'} = 1e-6;
    bless $self,$class;
    return $self;
}

sub functions()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ("dC_$n = ".$self->value('dC'),"fdC_$n = 1.0");
    push @l,("t_$n = ".$self->value('t'),"ft_$n = 1.0");
    push @l,"a_$n = ".$self->value('a');
    push @l,"b_$n = ".$self->value('b');
    push @l,'z_'.$n.'(x)= ((1 + ({0,1} * 2 * pi * x * t_'.$n.' * ft_'.$n.')**a_'.$n.')**b_'.$n.') / ( {0,1} * 2 * pi * x * (dC_'.$n.' * fdC_'.$n.'))';
    return join("\n",@l)."\n";
}

sub F($)
{
    my $self = shift;
    my $f = shift;
    return undef unless ($f && $f > 0);
    my $a = $self->value('a');
    my $b = $self->value('b');
    my $t = $self->value('t');
    my $c = $self->value('dC');
    return ((1 + (cplx(0,1) * 2 * pi() * $f * $t)**$a)**$b) / (cplx(0,1) * 2 * pi() * $f * $c);
}

sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    return ("fdC_$n","t_$n","a_$n","b_$n");
}

sub save()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ();
    push @l,"Final_dC_$n = dC_$n * fdC_$n";
    push @l,"Final_t_$n = t_$n * ft_$n";
    push @l,"Final_a_$n = a_$n";
    push @l,"Final_b_$n = b_$n";
    return join("\n",@l)."\n";
}


return 1;

