package Impedance::IMPCGI;
#$Id: IMPCGI.pm,v 1.55 2015/10/21 06:09:30 sofc Exp $
##########################################################################
#
#  Package Impedance::IMPCGI
#
# This module contains a number of utility functions for outputting properly formatted html code for
# user interface generation. Thus it mainly extends the CGI.pm module by Lincoln D. Stein.
# The module exports these functions in two groups.
#
# The :html group exports these functions:
#
# -- \begin{longtable}{ l p{10cm} }
# -- print_header($[\%]) & Prints the header information. Arguments: title. Any additional optional arguments (in the form of a hash) will be parsed along to the header() function supplied by CGI.pm. The functionautomatically appends a call to a javascript function logging users out after some time of no actions. \\
# -- print_end() & Prints the help button and ends the html output with the proper tag.  \\
# -- print_hidden() & Prints a number of hidden fields used to maintain state. \\
# -- logout() & Printys a logout button. \\
# -- action($) & Prints a hidden field with an action parameter with the specified value which can be used for program control flow.  \\
# -- EFA_start_html() & A wrapper for CGI::start_html. Any arguments (in the form of a hash) are passed to CGI::start_html. Automatically appends a reference to the javascript source file on the server. \\
# -- js_back() & Prints the javascript for gping backwards (uses the browser.back() fjavascrpt call). \\
# -- get_CGI_value($) & Retrieves the value of the specified CGI parameter (supplied by the web browser. \\
# -- get_CGI_value_clean($) & Same as get_CGI_value, but does pattern match on the retrieved value and only returns the part that matches. The pattern match is [{\backslash}w{\backslash}s{\backslash}.{\backslash},]*. This has the benefit of untainting the returned parameter value (For taint checks in perl and web access, refer Lincoln D. Steins book \textit{Official Guide to Programming with CGI.pm}) \\
# --\end{longtable}
#
#
# The :cgi group of functions include the following:
#
# --\begin{longtable}{ l p{10cm} }
# -- get_CGI_value($) & See above. \\
# -- get_CGI_value_clean($) &  See above. \\
# -- action($) & See above \\
# -- menu_button(@) & Prints a menu button. Arguments: name, value, style. The name will be the CGI parameter name, the value vill be the text on the button and the style is a style class name to use for displaying. \\
# -- create_menu_field & Prints the html tags to create a menu field.  \\
# -- top_nav_bar_start() & Prints the html tags to start the top navigation bar (table specifications etc.) \\
# -- top_no_button() & Prints a no action button (goes nowhere) in the top navigation bar. \\
# -- top_nav_bar_button() & prints a top navigation button. Arguments: File, name, value, style, [optional additional name, value and force triplets]. The file is the cgi-script to be called upon button press, the name,value and style arguments are passed to menu_button() and the additional optional arguments are used to initialise and print hidden html fields in the form of name-value pairs and a force argument (1 for force value, 0 for allow reuse of value). \\
# -- tab_newrow() & Prints a new row in the top navigation bar. \\
# -- top_js_return() & Prints a top navigation return button (uses the javascript printed by js_back(), see above) \\
# -- end_top_bar() & Prints the end of the top navigation bar. \\
# --\end{longtable}
#
# An additional function which is often used (but which is not exported by default) is the
# format_model($) function which parses a given impedance model and removes any empty serial or parallel connections as well as empty lines.
# The format_model function also removes any "\r" characters which may have been added by Microsoft based browsers.
# The function returns the resulting model.
#
##########################################################################
use strict;
use warnings;
use Impedance::Header;
use CGI qw(:standard :netscape);
use constant UPLOAD_DIR => '/tmp/';
use constant BUFF => 16_384;
use constant MAX_F_SIZE => 2 * 1_048_576; ### 2 MB ###
use constant MAX_DIR_SIZE => 100 * 1_048_576; ### 100 MB ###
use constant MAX_OPEN_TRIES => 100;
use Fcntl qw(:DEFAULT :flock);
$CGI::DISABLE_UPLOADS = 0;
$CGI::POST_MAX = MAX_F_SIZE;
require Exporter;

BEGIN {
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    @ISA = qw(Exporter);
    @EXPORT = qw($homedir $userdir);
    @EXPORT_OK = qw(top_nav_bar_button print_header print_end not_auth print_hidden action EFA_start_html js_back get_CGI_value get_CGI_value_clean js_back menu_button create_menu_field top_no_button tab_newrow action end_top_bar top_nav_bar_start top_js_return create_menu_field top_nav_bar_button_js);
    %EXPORT_TAGS = (
		    html => [ qw(print_header print_end print_hidden action EFA_start_html js_back get_CGI_value get_CGI_value_clean js_back create_menu_field) ],
		    cgi => [ qw(get_CGI_value get_CGI_value_clean menu_button create_menu_field top_no_button tab_newrow action end_top_bar top_nav_bar_start top_js_return top_nav_bar_button top_nav_bar_button_js) ],
                    );
}

my $msg = '';
our $name = $Impedance::Header::COPYRIGHT.' Impedance visualisation and analysis';

return 1;


sub action($)
#############################################################
### This function prints a hidden field with hame action ####
### and value equal to the argument                      ####
### Usage  action($value);                               ####
#############################################################
{
    print hidden(-name=>'action',-value=>shift,-force=>'1');
}

sub EFA_start_html
#####################################################################
### This function is a wrapper for start_html but automatically #####
### if the onload argument is already given the logout script   #####
### is just added to the functions which are to be run.         #####
#####################################################################
{
    my %h = @_;
    my $onload = 0;
    for my $key (keys %h)
    {
        if ($key =~ /-onload/i)
        {
            my $str = $h{$key};
            my $e = '';
            if ($str =~ /([\'\"])$/)
            {
                $e = $1;
                $str =~ s/[\'\"]$//;
            }
            $str = $str . ';'.$e;
            $h{$key} = $str;
            $onload = 1;
        }
    }
    $h{'-script'} = { -language => 'javascript', -src => '/analytic/analytic.js' };
    return CGI::start_html(%h);
}

sub print_header($)
#########################################################
### This function prints the header for the web pages ###
### Any second argument is used as the javascript     ###
### function to be called after the page loads        ###
#########################################################
{
    my $tit = shift;
    my $use_time = 0;
    my $res = header(-charset=>'utf-8');
    my %h = @_;
    my $onload = 0;
    for my $key (keys %h)
    {
        if ($key =~ /-onload/i)
        {
            my $str = $h{$key};
            my $e = '';
            if ($str =~ /([\'\"])$/)
            {
                $e = $1;
                $str =~ s/[\'\"]$//;
            }
            $str = $str . ';doAlertLogout();'.$e if ($use_time);
            $h{$key} = $str;
            $onload = 1;
        }
    }
    if ($use_time && !$onload)
    {
        $res .= EFA_start_html(%h,-title=>$tit,-style=>{src=>'/analytic/analytic.css'},-onLoad=>'doAlertLogout()');
    }
    else
    {
        $res .= EFA_start_html(%h,-title=>$tit,-style=>{src=>'/analytic/analytic.css'});
    }
    print $res;
    print "\n\n";
    print start_form(-name=>'login_info');
    print_hidden();
    print end_form();
    print "\n\n";
}

sub print_end()
#######################################
### Prints the end_html command #######
#######################################
{
    my $file = 'manual.pdf';
    $file = shift if (@_);
    print "\n\n";
    print p();
    print center(a({-href=>'javascript:load_help();'},'Help'));
    print <<END;

<script lanuage="javascript">
function load_help()
{
    var myWindow=window.open('/analytic/$file','windowRef','width=700,height=500,left=100,top=100,resizable=yes,scrollbars=yes');
    myWindow.location.href = '/analytic/$file';
    if (!myWindow.opener)
    {
        myWindow.opener = self;
        myWindow.focus();
    }
}

</script>

END
    print center(p('Copyright &copy '.$Impedance::Header::COPYYEAR.' '.$Impedance::Header::IMPLEMENTERS.', '.$Impedance::Header::COPYRIGHT));
    print end_html;
}

sub js_back
##############################################
### This function prints a javascript which ##
### takes the browser one step backwards    ##
##############################################
{
    print "\n",'<script>',"\n";
    print 'function back() {',"\n";
    print '  history.back();',"\n";
    print '  return false;',"\n";
    print '}',"\n";
    print '</script>',"\n";
}

sub print_hidden()
#########################################################################
### Prints the logininformation to be transfered to the next cgi call ###
### If called with arguments the username and password is used from   ###
### the passed arguemnts instead from CGI parameters                  ###
#########################################################################
{
    my $data = param('data');
    if (defined($data))
    {
	my @lines = split(/\n/,$data);
	foreach my $item (@lines)
	{
	    $item =~ s/[\s\t]+/ /g;
	    $item =~ s/^\s*//;
	    $item =~ s/E/e/g;
	    my @l = split(/\s+/,$item);
	    foreach my $i (@l)
	    {
		$i = sprintf("%.6g",$i);
	    }
	    $item = join(' ',@l);
	}
	$data = join("\n",@lines);
    }
    print hidden(-name=>'data',-value=>$data,-force=>1);
    my $model = &format_model(param('model'));
    print hidden(-name=>'model',-value=>$model,-force=>1);
    my $mode = param('mode');
    print hidden(-name=>'mode',-value=>$mode,-force=>1);
    my $minf = get_CGI_value_clean('minf');
    $minf = 1 unless ($minf);
    my $maxf = get_CGI_value_clean('maxf');
    $maxf = 10000 unless ($maxf);
    print hidden(-name=>'minf',-value=>$minf,force=>1);
    print hidden(-name=>'maxf',-value=>$maxf,force=>1);
    print hidden(-name=>'E_file');
    print hidden(-name=>'RFC_file');
}

sub format_model($)
{
    my $model = shift;
    if ($model)
    {
	$model =~ s/[\r\n]+/\n/g;
	### remove empty serial or paralel devices ###
	while ($model =~ /[\(\[][\s\t\n]*[\]\)]/)
	{
	    $model =~ s/[\(\[][\s\t\n]*[\]\)]//;
	}
	my @l = split(/\n+/,$model);
	my @r = ();
	### Remove empty lines ###
	foreach (@l)
	{
	    push @r,$_ if (/[\w\[\]\(\)]+/);
	}
	$model = join("\n",@r);
    }
    return $model;
}

sub top_nav_bar_start
##################################################
###Function creating the top_navigator_bar   #####
###                                          #####
##################################################
{
    print '<table /*border="2px"*/ class="top_table", width=100% >';
    print '<col width=20%", span="5">';
    print '<tr>';
    &create_menu_field;
    print start_form(-action=>'main.cgi');
    print submit(-class=>'s0',-name=>'Exit',-value=>'Exit');
    print end_form;
    print '</td>';
    &print_page();
}

sub print_page()
###################################################
### Function creating a top bar button for      ###
### printing the active page                    ###
###################################################
{
    print "\n<script lanuage=\"javascript\">\n";
    print "function printWindow() {\n";
    print "   var bV = parseInt(navigator.appVersion);\n";
    print "   if (bV >= 4) {\n";
    print "      window.print();\n";
    print "   }\n";
    print "}\n";
    print "</script>\n";
    print '<td>';
    print button(-name=>'Printpage',-value=>'Print page',-class=>'print',-onclick=>'printWindow()');
    print '</td>';
}


sub top_nav_bar_button($$@)
#######################################################################
### Prints a nav bar button which when pressed                     ####
### takes the user to the $file CGI-script and                     ####
### parses the action parameter $action to the                     ####
### script. For the 3 last arguments, see                          ####
### menu_button()                                                  ####
### Usage top_nav_bar_button($file,$action,$1,$2,$3,[opt args])    ####
### Arguments 1 to 3 are parced directly to                        ####
### menu_button().                                                 ####
### It is possible to specify parameters to be printed in hidden   ####
### fields by adding additional arguments, each additional         ####
### argument must be a comma separated list specifying (in order!) ####
### name, value and force (0/1). if the parameter is just to be    ####
### 'sticky, let the last two elements be blank, eks: search_val,, ####
### would print a hidden field with name 'search_val' and the      ####
### value would be whatever the previous field with name           ####
### 'searh_val' had.                                               ####
#######################################################################
{
    my $file = shift;
    my $action = shift;
    &create_menu_field;
    if ($file)
    {
        print start_form(-action=>$file,-class=>"form");
    }
    else
    {
        print start_form(-class=>"form");
    }
    &action($action) if ($action);
    my @args = ();
    push @args, shift @_;
    push @args, shift @_;
    push @args, shift @_;
    &menu_button(@args);
    &print_hidden();
    while (@_)
    {
        my @s = split(/\,/,shift @_);
        print hidden(-name=>$s[0],-value=>$s[1],-force=>$s[2]);
    }
    print end_form();
    print '</td>';
}

sub top_nav_bar_button_js($$@)
#######################################################################
### Prints a nav bar button which when pressed                     ####
### takes the user to the $file CGI-script and                     ####
### parses the action parameter $action to the                     ####
### script. For the 4 last arguments, see                          ####
### menu_button_js().                                              ####
### Usage top_nav_bar_button($file,$action,$1,$2,$3,[opt args])    ####
### Arguments 1 to 4 are parced directly to                        ####
### menu_button_js().                                              ####
### It is possible to specify parameters to be printed in hidden   ####
### fields by adding additional arguments, each additional         ####
### argument must be a comma separated list specifying (in order!) ####
### name, value and force (0/1). if the parameter is just to be    ####
### 'sticky, let the last two elements be blank, eks: search_val,, ####
### would print a hidden field with name 'search_val' and the      ####
### value would be whatever the previous field with name           ####
### 'searh_val' had.                                               ####
#######################################################################
{
    my $file = shift;
    my $action = shift;
    &create_menu_field;
    if ($file)
    {
        print start_form(-action=>$file,-class=>"form");
    }
    else
    {
        print start_form(-class=>"form");
    }
    &action($action) if ($action);
    my @args = ();
    push @args, shift @_;
    push @args, shift @_;
    push @args, shift @_;
    push @args, shift @_;
    &menu_button_js(@args);
    &print_hidden();
    while (@_)
    {
        my @s = split(/\,/,shift @_);
        print hidden(-name=>$s[0],-value=>$s[1],-force=>$s[2]);
    }
    print end_form;
    print '</td>';
}



sub top_no_button
#################################
### Prints a 'empty' button   ###
### in the top nav bar        ###
#################################
{
    &create_menu_field;
    print start_form;
    print button(-name=>'dummy',-value=>'',-class=>'none',-onclick=>'return false;');
    print end_form;
    print '</td>';
}

sub top_js_return
#############################################################
### Prints a return button  in the top navbar which uses  ###
### javascript to bring the user one step back            ###
### Note that the function requires that the function     ###
### &js_back() is used somewhere on the page!!!           ###
#############################################################
{
    &create_menu_field;
    print start_form(-class=>"form",-onsubmit=>"return back()");
    &print_hidden();
    &menu_button('Back','Back','back');
    print end_form();
    print '</td>';
}

sub tab_newrow
################################
### prints a html collumn    ###
### separator                ###
################################
{
    print '</tr><tr>';
}

sub end_top_bar
#################################
### Ends the top nav bar     ####
#################################
{
    print '</tr>';
    print '</table>';
    print start_form(-name=>'sofc');
    print_hidden();
    print hidden(-name=>'offset',-value=>0,-force=>1);
    print hidden(-name=>'orgoffset',-value=>0,-force=>1);
    print hidden(-name=>'length',-value=>0,-force=>1);
    print hidden(-name=>'orglength',-value=>0,-force=>1);
    print hidden(-name=>'text',-value=>'',-force=>1);
    print hidden(-name=>'fitlist',-value=>'',-force=>1);
    print hidden(-name=>'bgcolor');
    print hidden(-name=>'arcs');
    print hidden(-name=>'type');
    print hidden(-name=>'plotmode');
    print hidden(-name=>'admode',-id=>'admode');
    print hidden(-name=>'fitmode');
    print hidden(-name=>'filename');
    print hidden(-name=>'filedata');
    print hidden(-name=>'fit_res');
    print hidden(-name=>'allow_close',-value=>1,-force=>1);
    print hidden(-name=>'action',-value=>'dummy',-force=>1);
    print hidden(-name=>'fit_table');
    print hidden(-name=>'add_R_L');
    print hidden(-name=>'xmin',-id=>'xmin');
    print hidden(-name=>'xmax',-id=>'xmax');
    print hidden(-name=>'ymin',-id=>'ymin');
    print hidden(-name=>'ymax',-id=>'ymax');
    print hidden(-name=>'gp_text',-id=>'gp_text');
    print hidden(-name=>'scale_factor',-id=>'scale_factor');
    print end_form();
}


sub create_menu_field
###################################################
### This function prints a table field with a   ###
### specific height                             ###
###################################################
{
    print '<td height="30px", margin="0px", padding="0px">';
}

sub menu_button($$$)
{
    my $name = shift;
    my $value = shift;
    my $style = shift;
    print submit(-name=>$name,-class=>$style,-value=>$value);
}

sub menu_button_js($$$)
{
    my $name = shift;
    my $value = shift;
    my $style = shift;
    my $script = shift;
    print button(-name=>$name,-class=>$style,-value=>$value,-onclick=>$script);
}

sub clean($)
#######################################################
#### This function cleans a string for all special  ###
#### characters (only allowed ones are . and ,)     ###
#######################################################
{
    my $val = shift;
    my $ret = undef;
    if (defined($val) && $val =~ /([\w\s\.\,\-]*)/)
    {
        $ret = $1;
    }
    return $ret;
}

sub get_CGI_value_clean($)
##################################################
#### Wrapper for &clean and &get_CGI_value    ####
##################################################
{
    my $id = shift;
    my $val = &get_CGI_value($id);
    return &clean($val);
}

sub get_CGI_value($)
####################################################
#### This function gets the CGI parameter with   ###
#### $name, returns undef if the parameter is    ###
#### not defined in the input parameters         ###
#### Usage: get_CGI_value($name)                 ###
####################################################
{
    my $id = $_[0];
    my $val = undef;
    $val = param($id) if defined(param($id));
    return $val;
}



#### Functions for printing the impedance model ####
#### Including buttons etc                      ####

sub p_element($)
{
    my $element = shift;
    my @list = ();
    if ($element->type =~ /par|ser/i)
    {
        print '<td colspan=3>';
        push @list,(&p_complex($element));
        print '</td>';
    }
    else
    {
	my @tags = $element->tags;
        my $nr = @tags;
        print '<th rowspan='.$nr.'>';
        my $name = $element->name;
        print button(-name=>$name,-value=>$name,-class=>'s0',-onclick=>'delete_element(this)');
        print '</th>';
        my $first = 0;
	my @fit = $element->fit;
        for( my $x = 0; $x < @tags; $x++)
        {
	    my $t = $tags[$x];
            print '</tr><tr>' unless (!$first++);
            $name =~ /_(\d+)/;
            print '<th>';
            my $nr = $1;
	    my $id = $nr.'_fit_'.$fit[$x];
            if (param('mode') && param('mode') =~ /simul/i)
            {
                print $t;
            }
            else
            {
		push @list,$id;
		my $checked = 1;
		my $fitlist = get_CGI_value('fitlist');
		if ($fitlist)
		{
		    my @l = split(/\,/,$fitlist);
		    foreach my $f (@l)
		    {
			next unless ($f =~ /$id/);
			if ($f =~ /=(\d)$/)
			{
			    $checked = $1;
			}
			last;
		    }
		}
                print checkbox(-id=>$id,-name=>$id,-label=>$t,-value=>1,-checked=>$checked);
            }
            print '</th><td>';
            print textfield(-id=>'value_'.$t.'_'.$nr,-name=>$nr.'-'.$t,-value=>$element->value($t),-force=>1,-onchange=>'set_new_val(this)',-size=>5);
	    ### Print help button ###
	    my $string = $element->description($t);
	    $string =~ s/\'/\\\"/g;  ## Fix quotes ##
	    print button(-name=>'dummy_desc_'.$t,-value=>'?',-onclick=>"alert('$string')");
            print '</td>';
        }
    }
    return @list;
}

sub p_complex($)
{
    my $e = shift;
    return unless ($e);
    my @tags = $e->tags;
    my $nr = 0;
    my @list = ();
    foreach my $ee ($e->elements)
    {
        $nr += $e->element($ee)->tags;
    }
    $nr = 2 * $nr - 1;
    print '<table border=0 cellpadding=10>';
    print "<tr><th rowspan=$nr>";
    my $name = $e->name;
    my $label = $name;
    $label =~ s/\s+/\n/g;
    print hidden(-name=>'nr_'.$name,-id=>'nr_'.$name,-value=>($e->nr - 1),-force=>1);
    my $pl = $e->print_line;
    my @l = split("\n",$pl);
    print hidden(-name=>'length_'.$name,-id=>'length_'.$name,-value=>scalar(@l),-force=>1);
    print hidden(-name=>'model_'.$name,-id=>'model_'.$name,value=>$pl,-force=>1);
    print button(-name=>$name,-value=>$e->type."\n".$label,-class=>'c_model',-onclick=>'edit_complex(this)');
    print '</th>';
    my $first = 0;
    foreach my $ee ($e->elements)
    {
        print '<tr><td colspan=3><hr></td></tr><tr>' unless (!$first++);
        push @list,(&p_element($e->element($ee)));
	print '</tr>';
    }
    print '</table>';
    return @list;
}


##############################################
#### This function returns the size of the ###
#### specified directory                   ###
#### Usage:  $size = dir_size($dir);       ###
##############################################
sub dir_size
{
    my $dir = shift;
    my $dir_size = 0;
    opendir DIR, $dir or do {
        &errorlog("Unable to open $dir: $!");
        return undef;
    };
    foreach ( readdir DIR )
    {
	my $size = -s "$dir/$_";
        $dir_size += $size if defined($size);
    }
    return $dir_size;
}


sub file_upload($)
######################################################################
### This function manages file upload for the database             ###
### Usage: file_upload(parameter_name);                            ###
### where the parameter_name is the name of the upload field       ###
### and $dir is the directory where the data should end.           ###
### Make sure that the diretory exsist and is writable by          ###
### Apache!                                                        ###
### The function returns undef if errors is encountered            ###
### and returns the absoulute filename of the uploaded file        ###
### if successfull.                                                ###
######################################################################
{
    my $id = shift;
    my $dir = '/tmp/';
    my $prefix = '';
    $dir =~ /([\/\w\.\-]+)/;
    $dir = $1;
    if ((length($dir) < 5) || (length(param($id)) < 5))
    {
        die "No directory or filename attempted in function Impedance::IMPCGI::file_upload(): File: $id, dir: $dir\n";
    }
    my $vfile = '';
    if ( &dir_size($dir) + $ENV{CONTENT_LENGTH} > MAX_DIR_SIZE)
    {
        die  "Error: Upload dir '$dir' is full, please increase the file space!\n";
        return undef;
    }
    else
    {
        my $filen = param($id);
        $filen =~ s/\s+/_/g;
        chomp $filen;
        if ($filen =~ /(\w[\w\.\-\s]*)$/)
        {
            $filen = $prefix.$1;
            my $fh = upload($id);
            if (!$fh)
            {
                die  "Error: Undefined filehandle: \'$fh\' associated with file upload field \'$id\' with value \'$filen\' in Impedance::IMPCGI::file_upload(). No file uploaded!\n";
            }
            if ($filen !~ /\_\.\w+$/)
            {
                $filen =~ s/(\.\w+)$/\_0$1/;
            }
            until (sysopen OUTPUT, $dir. $filen, O_CREAT | O_RDWR | O_EXCL )
            {
                $filen =~ s/(\d*)(\.\w+)$/($1||0)+1 . $2/e;
                $1 >= MAX_OPEN_TRIES and do
                {
                    die "Error: Unable to get unique name $filen for file upload in Impedance::IMPCGI::file_upload().\n";
                };
            }
            my $buff = undef;
            $vfile = $dir.$filen;
            while (read($fh,$buff, BUFF))
            {
                print OUTPUT $buff;
            }
            close OUTPUT;
        }
        else
        {
            die  "Error: Invalid filename attempted for upload: $filen in Impedance::IMPCGI::file_upload().\n";
        }
    }
    return $vfile;
}
